import logging
from ruxit.api.base_plugin import BasePlugin
from ruxit.api.exceptions import ConfigException
from ruxit.api.data import MEAttribute
from dynatrace import Dynatrace
from models import MintMetric
import socket
import select
import requests
from datetime import datetime, timedelta
import os
from xml.etree import ElementTree
from xml.parsers.expat import ExpatError

logger = logging.getLogger(__name__)


class dbGangliaIngest(BasePlugin):
    def initialize(self, **kwargs):
        config = kwargs['config']
        self.dt_url = config.get("dt_api_url", None)
        if self.dt_url:
            self.dt_url = self.dt_url.strip("/")
        self.dt_token = config.get("dt_api_token", None)
        if not self.dt_token or not self.dt_url:
            raise ConfigException("Dynatrace API URL or Token cannot be empty")
        elif self.dt_url.endswith("/v1/api") or self.dt_url.endswith("/v2/api"):
            raise ConfigException("Please use base Dynatrace URL, see hint.")
        else:
            self.dt = Dynatrace(self.dt_url, self.dt_token)
        self.logger = logger
        self.host_first_seen_time = {}
        self.upsizing_timer = None
        self.downsizing_timer = None

    def query(self, **kwargs):
        config = kwargs['config']
        self.driver_host_id = None
        self.report_properties()
        self.config_metrics_capture(config)
        ganglia_data = self.read_xml_metrics()
        if ganglia_data:
            mint_lines = self.parse_xml_to_metrics(ganglia_data)
            for lines_subset in self.chunks(mint_lines, 1000):
                #self.logger.info(lines_subset)
                try:
                    json_response = self.ingest(lines_subset)
                    self.logger.info(f"DB Ganglia : {str(json_response)}")
                except requests.exceptions.ConnectionError as conn_error:
                    self.logger.exception(f"DB Ganglia : {conn_error}")
            mint_lines = self.cluster_event_metrics()
            for lines_subset in self.chunks(mint_lines, 1000):
                #self.logger.info(lines_subset)
                try:
                    json_response = self.ingest(lines_subset)
                    self.logger.info(f"DB Ganglia : {str(json_response)}")
                except requests.exceptions.ConnectionError as conn_error:
                    self.logger.exception(f"DB Ganglia : {conn_error}")
        else:
            self.logger.warn("DB Ganglia : No data returned from Ganglia API. This is okay if this server is not the driver node.")


    def chunks(self, elements, size):
        for i in range(0, len(elements), size):
            yield elements[i : i + size]      

    def check_capture_metric(self, metric_name):
        if metric_name.startswith("app"):
            app, metric_name = metric_name.split(".", 1)
        pieces = metric_name.split(".")
        metric_group = f"{pieces[0]}.{pieces[1]}"
        if metric_group in self.metric_startswith:
            if pieces[-1] in self.metric_endswith:
                return False
            else:
                return True
        else:
            return False

    def cluster_event_metrics(self):
        mint_lines = []
        dimensions = {}
        dimensions["cluster_id"] = self.friendly_cluster_id
        dimensions["cluster_name"] = self.friendly_cluster_name
        if self.driver_host_id:
            dimensions["dt.entity.host"] = self.driver_host_id
        databricks_url = None
        databricks_token = None
        try:
            with open("/etc/databricks_env", "r") as f:
                lines = f.readlines()
            for line in lines:
                if "DB_WS_URL" in line:
                    databricks_url = line.split("=")[-1].strip()
                    databricks_url = databricks_url.strip("/")
                if "DB_WS_TOKEN" in line:
                    databricks_token = line.split("=")[-1].strip()
        except Exception as e:
            self.logger.warn(f"DB Ganglia : Could not get Databricks Workspace URL or Token : {e}")
        
        if self.friendly_cluster_id and databricks_url and databricks_token:
            minute_ago = str((datetime.now().timestamp() * 1000) - 60000).split(".")[0]
            minute_ago = int(minute_ago)
            HEADERS = {"Authorization": f"Bearer {databricks_token}"}
            data_dict = {"cluster_id": self.friendly_cluster_id, 
                        "start_time": minute_ago,
                        "limit": 500
                        }
            try:
                response = requests.post(url=f"{databricks_url}/api/2.0/clusters/events", headers=HEADERS, json=data_dict)
                self.logger.info(response.json())
                events = response.json()["events"]
                for event in events:
                    if event["type"] == "RESIZING":
                        details = event["details"]
                        current = details['current_num_workers']
                        target = details['target_num_workers']
                        if target > current:
                            #Upsizing
                            self.logger.info("DB Ganglia : Cluster started upsizing")
                            self.upsizing_timer = datetime(1970, 1, 1) + timedelta(milliseconds=event['timestamp'])
                        elif target < current:
                            #Downsizing
                            pass
                    if event["type"] == "UPSIZE_COMPLETED":
                        if self.upsizing_timer:
                            self.logger.info("DB Ganglia : Cluster finished upsizing")
                            upsize_dt = datetime(1970, 1, 1) + timedelta(milliseconds=event['timestamp'])
                            upsizing_time = upsize_dt - self.upsizing_timer
                            upsizing_time_ms = str(upsizing_time.total_seconds() * 1000).split(".")[0]
                            self.upsizing_timer = None
                            mint_metric = MintMetric(key="cluster_upsizing_time", value=upsizing_time_ms, value_type="gauge", dimensions=dimensions, unit="MilliSecond")                     
                            mint_lines.append(mint_metric.to_mint())
                            mint_lines.append(mint_metric.to_mint_meta())
            except Exception as e:
                self.logger.warn(f"DB Ganglia : Could not parse Cluster Event Logs : {e}")
        else:
            self.logger.warn(f"DB Ganglia : Not collecting Cluster Events, missing cluster id: {self.friendly_cluster_id}, or databricks url: {databricks_url} or token: {databricks_token} ")      
        return mint_lines

    def config_metrics_capture(self, config):
        self.metric_endswith = []
        self.metric_startswith = []

        percentile_metric = config.get("percentile_metrics", False)
        stddev_metrics = config.get("stddev_metrics", False)

        if not percentile_metric:
            self.metric_endswith.append("p50") 
            self.metric_endswith.append("p75") 
            self.metric_endswith.append("p95") 
            self.metric_endswith.append("p98") 
            self.metric_endswith.append("p99") 
            self.metric_endswith.append("p999")
        if not stddev_metrics:
            self.metric_endswith.append("stddev")

        driver_blockmanager = config.get("driver_blockmanager", False)
        if driver_blockmanager:
            self.metric_startswith.append("driver.BlockManager")
        driver_codegenerator = config.get("driver_codegenerator", False)
        if driver_codegenerator:
            self.metric_startswith.append("driver.CodeGenerator")
        driver_dagscheduler = config.get("driver_dagscheduler", False)
        if driver_dagscheduler:
            self.metric_startswith.append("driver.DAGScheduler")
        driver_databricks = config.get("driver_databricks", False)
        if driver_databricks:
            self.metric_startswith.append("driver.Databricks")
        driver_hiveexternalcatalog = config.get("driver_hiveexternalcatalog", False)
        if driver_hiveexternalcatalog:
            self.metric_startswith.append("driver.HiveExternalCatalog")
        driver_executormetrics = config.get("driver_executormetrics", False)
        if driver_executormetrics:
            self.metric_startswith.append("driver.ExecutorMetrics")
        driver_jvmcpu = config.get("driver_jvmcpu", False)
        if driver_jvmcpu:
            self.metric_startswith.append("driver.JVMCPU")
        driver_livelistenerbus = config.get("driver_livelistenerbus", False)
        if driver_livelistenerbus:
            self.metric_startswith.append("driver.LiveListenerBus")

    def convert_units(self, metric_unit):
        if metric_unit == "%":
            return "Percent"
        elif metric_unit == "bytes/sec":
            return "BytePerSecond"
        elif "/sec" in metric_unit:
            return "PerSecond"
        elif metric_unit == "milliseconds":
            return "MilliSecond"
        elif metric_unit == "s":
            return "Second"
        elif metric_unit == "KB":
            return "KiloByte"
        elif metric_unit == "GB":
            return "GigaByte"
        elif metric_unit == "MHz":
            return "Hertz"
        else:
            return "Count"

    def find_cluster_metadata(self):
        self.friendly_cluster_id = None
        self.friendly_cluster_name = None
        try:
            with open("/databricks/common/conf/deploy.conf", "r") as f:
                lines = f.readlines()
            for line in lines:
                if "spark.databricks.clusterUsageTags.clusterName" in line:
                    parts = line.split("=")
                    santized = parts[1].strip()
                    self.friendly_cluster_name = santized.strip('"')
                elif "spark.databricks.clusterUsageTags.clusterId" in line:
                    parts = line.split("=")
                    santized = parts[1].strip()
                    self.friendly_cluster_id = santized.strip('"')
        except Exception as e:
            self.logger.exception(f"DB Ganglia : Error reading '/databricks/common/conf/deploy.conf' file : {e}")

    def find_dt_host_entities(self, grid_authority_id):
        self.dt_host_entities = {}
        try:
            for entity in self.dt.entities.list(f'type("HOST"),entityName("{grid_authority_id}")', time_from="now-5m"):
                full_entity = self.dt.entities.get(entity.entity_id)
                detected_name = full_entity.properties['detectedName']
                self.dt_host_entities[detected_name] = entity.entity_id
        except Exception as e:
            self.logger.exception(f"DB Ganglia : Could not retrieve Dynatrace Host Entities : {e}")

    def read_data_from_port(self, host, port, send=None):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect((host, port))
            r, w, x = select.select([sock], [], [], 2)
            if not r:
                sock.close()
                return
        except ConnectionRefusedError as e:
            self.logger.warn(f"DB Ganglia : {host}:{port} - Connection Refused Error")
            return
        except socket.error as e:
            self.logger.exception(f"DB Ganglia : Error connecting to {host}:{port} : {e}")
            return

        try:
            if send is not None: sock.send(send)
        except socket.error as e:
            self.logger.exception(f"DB Ganglia : Error connecting to {host}:{port} : {e}")
            return
        buffer = ""
        while True:
            try:
                data = sock.recv(8192)
            except socket.error as e:
                return

            if not data: break
            buffer += data.decode("ISO-8859-1")

        sock.close()
        return buffer

    def read_xml_metrics(self):
        xml_data = self.read_data_from_port("localhost", 8651)
        if xml_data:
            try:
                ganglia = ElementTree.XML(xml_data)
                return ganglia
            except UnicodeEncodeError:
                self.logger.warn('DB Ganglia : Could not parse XML data')
                return None
        else:
            return None

    def report_properties(self,):
        try:
            with open("/databricks/common/conf/deploy.conf", "r") as f:
                lines = f.readlines()
            for line in lines:
                if "spark.databricks.driverNodeTypeId" in line:
                    parts = line.split("=")
                    santized = parts[1].strip()
                    self.driver_type = santized.strip('"')
                    self.results_builder.report_property(key="Driver Type", value=self.driver_type, me_attribute=MEAttribute.CUSTOM_HOST_METADATA)
                elif "spark.databricks.workerNodeTypeId" in line:
                    parts = line.split("=")
                    santized = parts[1].strip()
                    self.worker_type = santized.strip('"')
                    self.results_builder.report_property(key="Worker(s) Type", value=self.worker_type, me_attribute=MEAttribute.CUSTOM_HOST_METADATA)
        except Exception as e:
            self.logger.exception(f"DB Ganglia : Error reading '/databricks/common/conf/deploy.conf' file : {e}")

    def parse_xml_to_metrics(self, ganglia):
        self.find_cluster_metadata()
        mint_lines = []
        try:
            for grid_elem in ganglia.findall("GRID"):
                grid_authority = grid_elem.get("AUTHORITY").split("/")[2]
                grid_authority_id = grid_authority.rsplit("-", 4)
                grid_authority_id = grid_authority_id[0]
                self.find_dt_host_entities(grid_authority_id)
                for cluster_elem in grid_elem.findall("CLUSTER"):
                    if not self.friendly_cluster_name:
                        self.friendly_cluster_name = cluster_elem.get("NAME")
                    alive_host_count = 0
                    dead_host_count = 0
                    for host_elem in cluster_elem.findall("HOST"):
                        host_name = host_elem.get("NAME")
                        dt_host_name = f"{grid_authority_id}-{host_name.replace('.','-')}"
                        if dt_host_name in self.dt_host_entities.keys():
                            host_id = self.dt_host_entities[dt_host_name]
                            json_response = self.tag_host(grid_authority, host_name, host_id)
                            self.logger.info(f"DB Ganglia : {str(json_response)}")
                        else:
                            host_id = None
                        host_tn = host_elem.get("TN")
                        host_tmax = host_elem.get("TMAX")
                        host_alive = True
                        if int(host_tn) > (2*int(host_tmax)):
                            dead_host_count += 1
                            host_alive = False
                        else:
                            alive_host_count += 1
                            host_alive = True
                        if host_alive:
                            for metric_elem in host_elem.findall("METRIC"):
                                metric_type = metric_elem.get("TYPE")
                                if metric_type != "string" and metric_type != "timestamp":
                                    try:
                                        metric_name = metric_elem.get("NAME")
                                        metric_name = metric_name.replace("$", "_")
                                        dimensions = {}
                                        dimensions["cluster_id"] = grid_authority_id
                                        dimensions["cluster_name"] = self.friendly_cluster_name
                                        if host_id:
                                            dimensions["dt.entity.host"] = host_id
                                        else:
                                            dimensions["host"] = dt_host_name
                                        if metric_name == "boottime":
                                            pieces = grid_authority.rsplit("-", 4)
                                            driver_ip = ".".join(pieces[1:])
                                            if driver_ip == host_name:
                                                dimensions["node_type"] = self.driver_type
                                            else:
                                                dimensions["node_type"] = self.worker_type
                                            now = datetime.now()
                                            if dt_host_name in self.host_first_seen_time.keys():
                                                first_seen = self.host_first_seen_time[dt_host_name]
                                                uptime = now - first_seen
                                                metric_value = int(str(uptime.total_seconds()).split(".")[0])
                                            else:
                                                self.host_first_seen_time[dt_host_name] = now
                                                metric_value = 0
                                            
                                            mint_metric = MintMetric(key="uptime", value=metric_value, value_type="gauge", dimensions=dimensions, unit="Second")
                                            mint_lines.append(mint_metric.to_mint())
                                            mint_lines.append(mint_metric.to_mint_meta())
                                        else:
                                            metric_value = metric_elem.get("VAL")
                                            metric_unit = metric_elem.get("UNITS").strip()
                                            dynatrace_unit = self.convert_units(metric_unit)
                                            if metric_name.endswith("_MB"):
                                                dynatrace_unit = "MegaByte"
                                            if dynatrace_unit == "Hertz":
                                                # MHz to Hz
                                                metric_value = str(int(metric_value) * 1000000)
                                            if "driver" in metric_name:
                                                if self.check_capture_metric(metric_name):
                                                    if metric_name.startswith("app"):
                                                        app, metric_name = metric_name.split(".", 1)
                                                        dimensions['application'] = app
                                                    mint_metric = MintMetric(key=metric_name, value=metric_value, value_type="gauge", dimensions=dimensions, unit=dynatrace_unit)
                                                    mint_lines.append(mint_metric.to_mint())
                                                    mint_lines.append(mint_metric.to_mint_meta())
                                            else:
                                                if metric_name.startswith("app"):
                                                    app, metric_name = metric_name.split(".", 1)
                                                    dimensions['application'] = app
                                                if "MajorGC" in metric_name or "MinorGC" in metric_name:
                                                    if "Time" in metric_name:
                                                        mint_metric = MintMetric(key=f"{metric_name}", value=metric_value, value_type="count", dimensions=dimensions, unit="MilliSecond")
                                                    else:
                                                        mint_metric = MintMetric(key=f"{metric_name}", value=metric_value, value_type="count", dimensions=dimensions, unit=dynatrace_unit)
                                                else:
                                                    mint_metric = MintMetric(key=metric_name, value=metric_value, value_type="gauge", dimensions=dimensions, unit=dynatrace_unit)
                                                mint_lines.append(mint_metric.to_mint())
                                                mint_lines.append(mint_metric.to_mint_meta())
                                    except Exception as e:
                                        self.logger.warn(f"DB Ganglia : Error gathering metric! : {e}")
                    
                    dimensions = {}
                    dimensions["cluster_id"] = grid_authority_id
                    dimensions["cluster_name"] = self.friendly_cluster_name
                    if self.driver_host_id:
                        dimensions["dt.entity.host"] = self.driver_host_id

                    mint_metric = MintMetric(key="alive_host_count", value=alive_host_count, value_type="gauge", dimensions=dimensions, unit="Count")                     
                    mint_lines.append(mint_metric.to_mint())
                    mint_metric = MintMetric(key="dead_host_count", value=dead_host_count, value_type="gauge", dimensions=dimensions, unit="Count")                     
                    mint_lines.append(mint_metric.to_mint())

        except Exception as e:
            self.logger.exception(f"DB Ganglia : Error parsing Ganglia XML Data : {e}")
        finally:
            return mint_lines

    def tag_host(self, grid_authority, host_name, host_id):
        pieces = grid_authority.rsplit("-", 4)
        driver_ip = ".".join(pieces[1:])
        json_dict = {}
        tag_list = []
        if driver_ip == host_name:
            self.driver_host_id = host_id
            tag = {}
            tag['key'] = "Databricks Driver Type"
            tag['value'] = self.driver_type
        else:
            tag = {}
            tag['key'] = "Databricks Worker Type"
            tag['value'] = self.worker_type
        tag_list.append(tag)
        json_dict["tags"] = tag_list
        return requests.post(url=self.dt_url.rstrip("/")+f'/api/v2/tags?entitySelector=entityId("{host_id}")', 
                             json=json_dict, 
                             headers={"Content-Type": "application/json; charset=utf-8", 
                                      "Authorization": f"Api-Token {self.dt_token}"}
                             ).json()       


    def ingest(self, lines):
        lines = "\n".join(lines).encode("utf-8")
        try:
            return requests.post(url="http://localhost:14499/metrics/ingest", 
                                data=lines, 
                                headers={"Content-Type": "text/plain; charset=utf-8",
                                        "eec-enrichment": "false"}
                                ).json()
        except requests.exceptions.ConnectionError as e:
            raise ConfigException("Please ensure OneAgent EEC & HTTP Metric Ingest are enabled.")
        except Exception as e:
            self.logger.exception(f"Metric Ingest Issue : {e}")